package com.hello2morrow.sonargraph.languageprovider.csharp.controller.settings;


import org.slf4j.Logger
import org.slf4j.LoggerFactory

import com.hello2morrow.foundation.file.FileUtility
import com.hello2morrow.foundation.utilities.IOMessageCause
import com.hello2morrow.foundation.utilities.OperationResult
import com.hello2morrow.foundation.utilities.OperationResultWithOutcome
import com.hello2morrow.foundation.utilities.Platform
import com.hello2morrow.foundation.utilities.StringUtility
import com.hello2morrow.sonargraph.core.controller.system.MSBuildExecutor
import com.hello2morrow.sonargraph.core.model.system.settings.IVisualStudioProjectType
import com.hello2morrow.sonargraph.core.model.system.settings.SettingScope
import com.hello2morrow.sonargraph.core.model.system.settings.SolutionFileInfo
import com.hello2morrow.sonargraph.core.persistence.system.settings.VisualStudioProjectType
import com.hello2morrow.sonargraph.languageprovider.csharp.foundation.common.CSharpCauses
import com.hello2morrow.sonargraph.languageprovider.csharp.model.path.CSharpFileType
import com.hello2morrow.sonargraph.languageprovider.csharp.model.settings.CSharpAssemblyReferenceInfo
import com.hello2morrow.sonargraph.languageprovider.csharp.model.settings.CSharpProjectReferenceInfo
import com.hello2morrow.sonargraph.languageprovider.csharp.model.settings.VisualStudioDeveloperShellInfo
import com.hello2morrow.sonargraph.languageprovider.csharp.model.settings.VsProjectImportModuleDto
import com.hello2morrow.sonargraph.languageprovider.csharp.persistence.visualstudio.CSharpVisualStudioFileProcessor

import de.schlichtherle.truezip.file.TFile

public class CSharpMSBuildExecutor extends MSBuildExecutor
{
    private static final Logger LOGGER = LoggerFactory.getLogger(CSharpMSBuildExecutor.class);

    private static final String START_TOKEN = "******* start *******";
    private static final String END_TOKEN = "******* end *******";
    private static final String CONFIGURATION_TOKEN = "Configuration:";
    private static final String PLATFORM_TOKEN = "Platform:";
    private static final String ASSEMBLY_NAME_TOKEN = "Assembly Name:";
    private static final String TARGET_FRAMEWORK_VERSION_TOKEN = "TargetFrameworkVersion:";
    private static final String TARGET_FRAMEWORK_PROFILE_TOKEN = "TargetFrameworkProfile:";
    private static final String PROJECT_TYPE_GUIDS_TOKEN = "ProjectTypeGuids:"
    private static final String FILES_TOKEN = "Compile Files:";
    private static final String CONSTANTS_TOKEN = "DefineConstants:";
    private static final String ASSEMBLIES_TOKEN = "Assemblies:";
    private static final String ASSEMBLY_HINTPATH_TOKEN = "Assembly HintPaths:";
    private static final String ASSEMBLY_ALIASES_TOKEN = "Assembly Aliases:";
    private static final String ASSEMBLY_FILE_TOKEN = "Assembly File:";
    private static final String REQUIRED_TARGET_FRAMEWORK_TOKEN = "RequiredTargetFrameworks:";
    private static final String PROJECT_REFERENCE_TOKEN = "Project Reference:";
    private static final String PROJECT_REFERENCE_NAME_TOKEN = "Project Reference Name:";
    private static final String HAS_SHARED_ITEMS_TOKEN = "Has Shared Items:";
    private static final String GET_REFERENCE_ASSEMBLY_PATHS_TOKEN = "GetReferenceAssemblyPaths:";

    private TFile m_executable;
    private VisualStudioDeveloperShellInfo m_developerShellInfo;
    private final boolean m_isSDKProject;

    private static final String STANDARD_PROJECT_TARGET = "ResolveAssemblyReferences";
    private static final String SDK_PROJECT_TARGET = "ResolveReferences";

    public CSharpMSBuildExecutor(TFile executable, TFile projectFile, TFile solutionDir, List<IVisualStudioProjectType> projectTypes)
    {
        super(projectFile, solutionDir);

        assert executable != null : "Parameter 'executable' of method 'CSharpMSBuildExecutor' must not be null";
        m_executable = executable;
        m_isSDKProject = isSdkProject(projectTypes);
    }
    
    public void setDeveloperShellInfo(VisualStudioDeveloperShellInfo developerShellInfo)
    {
        assert developerShellInfo != null : "Parameter 'developerShellInfo' of method 'setDeveloperShellInfo' must not be null";
        m_developerShellInfo = developerShellInfo;
    }

    private boolean isSdkProject(List<IVisualStudioProjectType> projectTypes)
    {
        return (projectTypes.contains(VisualStudioProjectType.STORE_APP_WINDOWS_PHONE_8_1)
                || projectTypes.contains(VisualStudioProjectType.STORE_APP_WINDOWS_PHONE_8_1_SILVERLIGHT)
                ||projectTypes.contains(VisualStudioProjectType.STORE_APP_WINDOWS_8_1)
                || projectTypes.contains(VisualStudioProjectType.STORE_APP_UNIVERSAL)
                || projectTypes.contains(VisualStudioProjectType.STORE_APP_PORTABLE_UNIVERSAL));
    }


    protected String getTemplateName()
    {
        return "msbuild.template";
    }

    protected String getWrapperFileName()
    {
        return "h2m" + CSharpFileType.PROJECT_FILE.getDefaultExtension();
    }

    protected Map<String, String> getProperties()
    {
        Map<String, String> props = super.getProperties();
        props.put("target", m_isSDKProject ? SDK_PROJECT_TARGET : STANDARD_PROJECT_TARGET);
        props.put("isWindows", !Platform.isOperatingSystemUnixBased());
        return props;
    }

    public OperationResultWithOutcome<VsProjectImportModuleDto> executeBuild(SolutionFileInfo.ConfigurationAndPlatform configurationAndPlatform, String toolsVersion)
    {
        assert configurationAndPlatform != null : "Parameter 'configurationAndPlatform' of method 'executeBuild' must not be null";
        assert toolsVersion != null && toolsVersion.length() > 0 : "Parameter 'toolsVersion' of method 'executeBuild' must not be empty";

        TFile wrapper = writeProjectWrapper();
        List<String> cmd = new ArrayList<>();
        OperationResultWithOutcome<VsProjectImportModuleDto> importResult = new OperationResultWithOutcome<VsProjectImportModuleDto>("Processing C# project file '" + getProjectFile().getName()+ "'");

        setExecutable(cmd, importResult);

        cmd.add("/nologo");
        cmd.add("/t:h2moptions");

        if (configurationAndPlatform != null)
        {
            cmd.add("/p:Configuration=$configurationAndPlatform.configuration");
            cmd.add("/p:Platform=$configurationAndPlatform.platform");
        }

        if (getSolutionDir() != null)
        {
            String path = getSolutionDir().getNormalizedAbsolutePath();
            cmd.add("/p:SolutionDir=$path");
        }

        //IK: Specifies the version of the Toolset to use to build the project
        //Must be at least version 4.0. Earlier versions don't contain GetReferenceAssemblyPaths
        //See SGNG-2706.
        cmd.add("/tv:" + toolsVersion);

        //Uncomment this if you want to generate a file containing the expanded preprocessed project file
        //cmd.add("/pp:" + new TFile("output").getNormalizedAbsolutePath());

        cmd.add(wrapper.getNormalizedAbsolutePath());

        String projectFilePath = getProjectFile().getNormalizedAbsolutePath();
        LOGGER.info("Executing '{}' for project file '{}'", cmd, projectFilePath);
        def writer = new StringWriter();
        int result = -1;
        try
        {
            def proc = cmd.execute();
            writer << proc.in;
            writer << proc.err;
            result = proc.waitFor();
        }
        catch (IOException ex)
        {
            importResult.addError(CSharpCauses.VISUAL_STUDIO_PROJECT_FILE_PROCESSING_FAILED, ex, "Failed to execute C# build with executable '" + m_executable.getAbsolutePath() + "'. ");
        }

        try
        {
            wrapper.rm()
        }
        catch(IOException ex)
        {
            LOGGER.error("Failed to delete file '{}'", wrapper.getAbsolutePath());
        }

        if (importResult.isSuccess())
        {
            String content = writer.toString();
            if (result == 0 && !content.contains("Error in script usage."))
            {
                importResult.setOutcome(processData(content, importResult));
            }
            else
            {
                importResult.addError(CSharpCauses.VISUAL_STUDIO_PROJECT_FILE_PROCESSING_FAILED, "Failed to process C# project file: $content");
            }
        }

        return importResult;
    }

    private void setExecutable(List<String> cmd, OperationResult result)
    {
        if (m_developerShellInfo != null && m_developerShellInfo.isActive())
        {
            TFile developerShellExecutable = (m_developerShellInfo.getExecutablePath() == null || m_developerShellInfo.getExecutablePath().isEmpty())
                    ? null : new TFile(m_developerShellInfo.getExecutablePath());
            if (developerShellExecutable == null)
            {
                cmd.add(m_executable.getNormalizedAbsolutePath());
                return;
            }

            if (!developerShellExecutable.exists())
            {
                result.addError(IOMessageCause.NOT_A_FILE, "Executable for Visual Studio Developer Prompt does not exist: " + m_developerShellInfo.getExecutablePath());
                return;
            }

            if (!developerShellExecutable.canExecute())
            {
                result.addError(IOMessageCause.NO_PERMISSION, "Cannot run exectuable for Visual Studio Developer Prompt: " + m_developerShellInfo.getExecutablePath());
                return;
            }

            cmd.add(developerShellExecutable.getNormalizedAbsolutePath());
            if (!m_developerShellInfo.getArgs().isEmpty())
            {
                for (String next : m_developerShellInfo.getArgs())
                {
                    cmd.add(next);
                }
            }

            cmd.add("&&");
            if (m_developerShellInfo.useMsBuildExecutable())
            {
                cmd.add(m_executable.getNormalizedAbsolutePath());
            }
            else
            {
                cmd.add(m_executable.getName());
            }

        }
        else
        {
            cmd.add(m_executable.getNormalizedAbsolutePath());
        }
    }

    private VsProjectImportModuleDto processData(String content, OperationResult result)
    {
        assert content != null : "parameter 'content' of method processData() must not be null";
        assert result != null : "parameter 'result' of method processData() must not be null";

        VsProjectImportModuleDto dto = new VsProjectImportModuleDto(getProjectFile(), getSolutionDir(), m_isSDKProject);
        boolean isStarted = false;
        String configuration = null;
        String platform = null;
        List<String> assemblies;
        List<String> assemblyHintPaths;
        List<String> aliases;
        List<String> targetFrameworks;
        List<String> projectReferences;
        List<String> projectReferenceNames;
        List<String> referenceAssemblyPaths;

        Map<String, String> assemblyNameToPathMap = new HashMap<>();

        for (String untrimmedLine : StringUtility.multiLineStringToList(content))
        {
            String line = untrimmedLine.trim();
            if (line.equals(START_TOKEN))
            {
                isStarted = true;
                continue;
            }
            if (line.equals(END_TOKEN))
            {
                break;
            }

            if (!isStarted)
            {
                continue;
            }

            if (line.startsWith(CONFIGURATION_TOKEN))
            {
                configuration = line.substring(CONFIGURATION_TOKEN.length());
                continue;
            }

            if (line.startsWith(PLATFORM_TOKEN))
            {
                platform = line.substring(PLATFORM_TOKEN.length());
                continue;
            }

            if (line.startsWith(ASSEMBLY_NAME_TOKEN))
            {
                dto.setAssemblyNameOfModule(line.substring(ASSEMBLY_NAME_TOKEN.length()));
                continue;
            }


            if (line.startsWith(TARGET_FRAMEWORK_VERSION_TOKEN))
            {
                String frameworkVersion = line.substring(TARGET_FRAMEWORK_VERSION_TOKEN.length());
                if (frameworkVersion.length() > 0)
                {
                    dto.setTargetFrameworkVersion(frameworkVersion);
                }
                continue;
            }

            if (line.startsWith(TARGET_FRAMEWORK_PROFILE_TOKEN))
            {
                String profileVersion = line.substring(TARGET_FRAMEWORK_PROFILE_TOKEN.length());
                if (profileVersion.length() > 0)
                {
                    dto.setTargetFrameworkProfile(profileVersion);
                }
                continue;
            }

            if (line.startsWith(PROJECT_TYPE_GUIDS_TOKEN))
            {
                String projectTypeGuids = line.substring(PROJECT_TYPE_GUIDS_TOKEN.length());
                if (projectTypeGuids.length() > 0)
                {
                    dto.setProjectTypeGuids(projectTypeGuids);
                }
                continue;
            }

            if (line.startsWith(FILES_TOKEN))
            {
                String files = line.substring(FILES_TOKEN.length());
                for (String fileName : StringUtility.split(files, ";"))
                {
                    String filePath = fileName.trim().replace('\\', '/');
                    dto.addCSharpFile(filePath);
                }
                continue;
            }

            if (line.startsWith(CONSTANTS_TOKEN))
            {
                String constants = line.substring(CONSTANTS_TOKEN.length());
                dto.setPreprocessorSymbolsText(new SolutionFileInfo.ConfigurationAndPlatform(configuration, platform), constants);
                continue;
            }

            if (line.startsWith(ASSEMBLIES_TOKEN))
            {
                String assemblyNames = line.substring(ASSEMBLIES_TOKEN.length());
                assemblies = (assemblyNames.length() > 0) ? Arrays.asList(assemblyNames.split(";", -1)) : Collections.emptyList();
                continue;
            }

            if (line.startsWith(ASSEMBLY_HINTPATH_TOKEN))
            {
                String assemblyHintPathString = line.substring(ASSEMBLY_HINTPATH_TOKEN.length());
                assemblyHintPaths = (assemblyHintPathString.length() > 0) ? Arrays.asList(assemblyHintPathString.split(";", -1)) : Collections.emptyList();
                continue;
            }

            if (line.startsWith(ASSEMBLY_ALIASES_TOKEN))
            {
                String assemblyAliases = line.substring(ASSEMBLY_ALIASES_TOKEN.length());
                aliases = Arrays.asList(assemblyAliases.split(";", -1));
                continue;
            }

            if (line.startsWith(REQUIRED_TARGET_FRAMEWORK_TOKEN))
            {
                String assemblyTargetFrameworks = line.substring(REQUIRED_TARGET_FRAMEWORK_TOKEN.length());
                targetFrameworks = Arrays.asList(assemblyTargetFrameworks.split(";", -1));
                continue;
            }

            if (line.startsWith(PROJECT_REFERENCE_TOKEN))
            {
                String projectReferencesString = line.substring(PROJECT_REFERENCE_TOKEN.length());
                projectReferences = Arrays.asList(projectReferencesString.split(";", -1));
                continue;
            }

            if (line.startsWith(PROJECT_REFERENCE_NAME_TOKEN))
            {
                String projectReferenceNamesString = line.substring(PROJECT_REFERENCE_NAME_TOKEN.length());
                projectReferenceNames = Arrays.asList(projectReferenceNamesString.split(";", -1));
                continue;
            }

            if (line.startsWith(HAS_SHARED_ITEMS_TOKEN))
            {
                String hasSharedItems = line.substring(HAS_SHARED_ITEMS_TOKEN.length());
                dto.setHasSharedItems(Boolean.parseBoolean(hasSharedItems));
            }

            if (line.startsWith(ASSEMBLY_FILE_TOKEN))
            {
                String fileNameAndPath = line.substring(ASSEMBLY_FILE_TOKEN.length());
                String[] parts = fileNameAndPath.split("\\|\\|", -1);
                if (parts.length == 2)
                {
                    String assemblyName = parts[0].trim();
                    String assemblyPath = parts[1].trim();
                    if (assemblyName.length() > 0 && assemblyPath.length() > 0)
                    {
                        assemblyNameToPathMap.put(assemblyName.toLowerCase(), assemblyPath)
                        LOGGER.debug("Determined assembly path -> " + assemblyPath)
                    }
                }
                else
                {
                    LOGGER.warn("line does not contain name and path of assembly: " + line);
                }
            }

            if (line.startsWith(GET_REFERENCE_ASSEMBLY_PATHS_TOKEN))
            {
                String assemblyPaths = line.substring(GET_REFERENCE_ASSEMBLY_PATHS_TOKEN.length());
                if (assemblyPaths.length() > 0)
                {
                    referenceAssemblyPaths = new ArrayList<>();
                    for (String nextPath : Arrays.asList(assemblyPaths.split(";")))
                    {
                        if (nextPath.endsWith("/") || nextPath.endsWith("\\"))
                        {
                            referenceAssemblyPaths.add(nextPath.substring(0, nextPath.length() - 1));
                        }
                    }
                }
            }
        }
        dto.setDefaultConfiguration(configuration);
        dto.setDefaultPlatform(platform);
        SolutionFileInfo.ConfigurationAndPlatform configAndPlatform = new SolutionFileInfo.ConfigurationAndPlatform(configuration, platform);
        if (!dto.getConfigurations().containsKey(configAndPlatform))
        {
            dto.addConfiguration(configAndPlatform);
        }
        processAssemblies(dto, assemblies, assemblyHintPaths, aliases, assemblyNameToPathMap, targetFrameworks, result);
        processProjectReferences(dto, projectReferences, projectReferenceNames, result);
        if (referenceAssemblyPaths != null && referenceAssemblyPaths.size() > 0)
        {
            dto.setReferenceAssemblyPaths(referenceAssemblyPaths);
        }
        return dto;
    }

    private void processAssemblies(VsProjectImportModuleDto dto, List<String> assemblies, List<String> assemblyHintPaths, List<String> aliases, Map<String, String> assemblyNameToPathMap, List<String> targetFrameworks, OperationResult result)
    {
        Map<String, String> additionalAssembliesMap = new HashMap<>(assemblyNameToPathMap);
        for (int i = 0; i < assemblies.size(); i++)
        {
            String assemblyName = assemblies.get(i);

            if (assemblyName.trim().length() == 0)
            {
                continue;
            }

            CSharpAssemblyReferenceInfo info = null;

            if (assemblyName.contains(","))
            {
                String[] parts = assemblyName.split(",", -1);
                String name = parts[0].trim();
                info = new CSharpAssemblyReferenceInfo(SettingScope.MODULE, name, null);
                TFile assemblyFile = getAssemblyFileAndRemoveFromMap(name, additionalAssembliesMap);
                info.setLocatedByBuildExecutor(assemblyFile != null);
                if (assemblyFile != null)
                {
                    info.setFileDeterminedByBuildExecutor(assemblyFile);
                }

                Map<String, String> assemblyAttributes = CSharpVisualStudioFileProcessor.convertAssemblyAttributesToMap(parts);
                String version = assemblyAttributes.get(CSharpVisualStudioFileProcessor.ASSEMBLY_VERSION);
                if (version != null)
                {
                    info.setVersion(version);
                }

                String culture = assemblyAttributes.get(CSharpVisualStudioFileProcessor.ASSEMBLY_CULTURE);
                if (culture != null)
                {
                    info.setCulture(culture);
                }

                String publicKey = assemblyAttributes.get(CSharpVisualStudioFileProcessor.ASSEMBLY_PUBLIC_KEY_TOKEN);
                if (publicKey != null)
                {
                    info.setPublicKey(publicKey);
                }

                String architecture = assemblyAttributes.get(CSharpVisualStudioFileProcessor.ASSEMBLY_PROCESSOR_ARCHITECTURE);
                if (architecture != null)
                {
                    info.setProcessorArchitecture(architecture);
                }
            }
            else
            {
                info = new CSharpAssemblyReferenceInfo(SettingScope.MODULE, assemblyName, null);
                TFile assemblyFile = getAssemblyFileAndRemoveFromMap(assemblyName, additionalAssembliesMap);
                info.setLocatedByBuildExecutor(assemblyFile != null);
                if (assemblyFile != null)
                {
                    info.setFileDeterminedByBuildExecutor(assemblyFile);
                }
            }

            TFile hintPathFile = null;
            if (assemblyHintPaths.size() > 0 && assemblyHintPaths.size() > i)
            {
                String hintPath = assemblyHintPaths.get(i);

                if (hintPath != null && hintPath.length() > 0)
                {
                    hintPath = FileUtility.convertPathToUniversalForm(hintPath);
                    if (FileUtility.isRelativePath(hintPath))
                    {
                        hintPathFile = new TFile(getProjectFile().getParentFile(), hintPath).getNormalizedAbsoluteFile();
                    }
                    else
                    {
                        hintPathFile = new TFile(hintPath);
                    }
                    info.setHintPath(hintPathFile);
                }
            }

            List<String> currentAliases;
            if (aliases.size() > 0 && aliases.size() > i)
            {
                String assemblyAliases = aliases.get(i);
                if (assemblyAliases != null && assemblyAliases.length() > 0)
                {
                    currentAliases = StringUtility.split(assemblyAliases, CSharpVisualStudioFileProcessor.EXTERN_ALIASES_SEPARATOR);
                    info.setExternAliases(currentAliases);
                }
            }

            if (targetFrameworks.size() > 0 && targetFrameworks.size() > i)
            {
                String targetFramework = targetFrameworks.get(i);
                if (targetFramework != null)
                {
                    info.setRequiredTargetFramework(targetFramework);
                }
            }

            if (!dto.addCSharpAssemblyReference(info))
            {
                result.addWarning(CSharpCauses.DUPLICATE_ASSEMBLY, "Duplicate reference to assembly '" + info.getName() + "' is ignored");
            }
        }

        //Process the left-overs, i.e. assemblies that are not explicitly referenced
        for (Map.Entry next : additionalAssembliesMap.entrySet())
        {
            TFile file = new TFile(next.getValue());
            String assemblyName = FileUtility.getFileNameWithoutExtension(file);
            CSharpAssemblyReferenceInfo info = new CSharpAssemblyReferenceInfo(SettingScope.MODULE, assemblyName, null);
            info.setVirtual(true);
            info.setLocatedByBuildExecutor(true);
            info.setFileDeterminedByBuildExecutor(file);
            if (!dto.addCSharpAssemblyReference(info))
            {
                result.addWarning(CSharpCauses.DUPLICATE_ASSEMBLY, "Duplicate reference to assembly '" + info.getName() + "' is ignored");
            }
            else
            {
                LOGGER.debug("Added additional (virtual) assembly: " + info.getName());
            }
        }
    }

    private TFile getAssemblyFileAndRemoveFromMap(String assemblyName, Map<String, String> additionalAssembliesMap)
    {
        assert assemblyName != null && assemblyName.trim().length() > 0 : "Parameter 'assemblyName' of method 'getAssemblyFile' must not be empty";
        assert additionalAssembliesMap != null  : "Parameter 'additionalAssembliesMap' of method 'getAssemblyFile' must not be null";

        TFile hintPathFile = null;
        String key = assemblyName.toLowerCase();
        String path = additionalAssembliesMap.get(key);
        if (path != null)
        {
            hintPathFile = new TFile(path);
            additionalAssembliesMap.remove(key);
        }
        return hintPathFile;
    }

    private void processProjectReferences(VsProjectImportModuleDto dto, List<String> projectReferences, List<String> projectReferenceNames, OperationResult result)
    {
        assert projectReferences != null : "Parameter 'projectReferences' of method 'processProjectReferences' must not be null";
        assert projectReferenceNames != null : "Parameter 'projectReferenceNames' of method 'processProjectReferences' must not be null";

        for (int i = 0; i < projectReferences.size(); i++)
        {
            String path = projectReferences.get(i);
            if (path == null || path.length() == 0)
            {
                continue;
            }

            if (path.startsWith("*Undefined*"))
            {
                result.addError(CSharpCauses.UNABLE_TO_RESOLVE_REFERENCED_PROJECT, "Unresolved project reference: " + path);
                continue;
            }

            path = FileUtility.convertPathToUniversalForm(path);
            TFile referenceFile = null;
            if (FileUtility.isRelativePath(path))
            {
                referenceFile = new TFile(getProjectFile().getParentFile(), path);
            }
            else
            {
                referenceFile = new TFile(path);
            }
            dto.addCSharpProjectReference(new CSharpProjectReferenceInfo(projectReferenceNames.get(i), referenceFile));
        }
    }
}